import os
from os.path import join as pjoin
import os.path
from refnx.reduce.platypusnexus import calculate_wavelength_bins
from refnx.reduce.reduce import PolarisationEfficiency
import warnings
import pytest

from numpy.testing import assert_equal, assert_allclose
import xml.etree.ElementTree as ET

from refnx.reduce import (
    reduce_stitch,
    PlatypusReduce,
    ReductionOptions,
    SpatzReduce,
    SpinSet,
    PolarisedReduce,
)
from refnx.dataset import ReflectDataset


class TestPlatypusReduce:
    @pytest.mark.usefixtures("no_data_directory")
    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir, data_directory):
        self.pth = pjoin(data_directory, "reduce")

        self.cwd = os.getcwd()
        self.tmpdir = tmpdir.strpath
        os.chdir(self.tmpdir)
        return 0

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_smoke(self):
        # a quick smoke test to check that the reduction can occur
        # warnings filter for pixel size
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            a, fname = reduce_stitch(
                [708, 709, 710],
                [711, 711, 711],
                data_folder=self.pth,
                reduction_options={"rebin_percent": 2},
            )
            a.save("test1.dat")
            assert os.path.isfile("./test1.dat")

            # reduce_stitch should take a ReductionOptions dict
            opts = ReductionOptions()
            opts["rebin_percent"] = 2

            a2, fname = reduce_stitch(
                [708, 709, 710],
                [711, 711, 711],
                data_folder=self.pth,
                reduction_options=[opts] * 3,
            )
            a2.save("test2.dat")
            assert os.path.isfile("./test2.dat")
            assert_allclose(a.y, a2.y)

    def test_reduction_method(self):

        # a quick smoke test to check that the reduction can occur
        # warnings filter for pixel size
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            a = PlatypusReduce("PLP0000711.nx.hdf", data_folder=self.pth)

            # try reduction with the reduce method
            a.reduce(
                "PLP0000708.nx.hdf",
                data_folder=self.pth,
                rebin_percent=4,
            )

            # try reduction with the __call__ method
            a(
                "PLP0000708.nx.hdf",
                data_folder=self.pth,
                rebin_percent=4,
            )

            # this should also have saved a couple of files in the current
            # directory
            assert os.path.isfile("./PLP0000708_0.dat")
            assert os.path.isfile("./PLP0000708_0.xml")

            # can we read the file
            ReflectDataset("./PLP0000708_0.dat")

            # try writing offspecular data
            a.write_offspecular("offspec.xml", 0)

    def test_free_liquids(self):
        # smoke test for free liquids
        a0 = PlatypusReduce("PLP0038418.nx.hdf", data_folder=self.pth)
        a1 = PlatypusReduce("PLP0038417.nx.hdf", data_folder=self.pth)

        # try reduction with the reduce method
        d0, r0 = a0.reduce(
            "PLP0038420.nx.hdf", data_folder=self.pth, rebin_percent=4
        )
        d1, r1 = a1.reduce(
            "PLP0038421.nx.hdf", data_folder=self.pth, rebin_percent=4
        )

    def test_event_reduction(self):
        # check that eventmode reduction can occur, and that there are the
        # correct number of datasets produced.

        # warnings filter for pixel size
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            a = PlatypusReduce(pjoin(self.pth, "PLP0011613.nx.hdf"))

            a.reduce(
                pjoin(self.pth, "PLP0011641.nx.hdf"),
                integrate=0,
                rebin_percent=2,
                eventmode=[0, 900, 1800],
            )

            assert_equal(a.y.shape[0], 2)

            # check that two datasets are written out.
            assert os.path.isfile("PLP0011641_0.dat")
            assert os.path.isfile("PLP0011641_1.dat")

            # check that the resolutions are pretty much the same
            assert_allclose(
                a.x_err[0] / a.x[0], a.x_err[1] / a.x[1], atol=0.001
            )

            # check that the (right?) timestamps are written into the datafile
            tree = ET.parse(pjoin(os.getcwd(), "PLP0011641_1.xml"))
            tree.find(".//REFentry").attrib["time"]
            # TODO, timestamp is created in the local time stamp of the testing
            # machine. The following works if reduced with a computer in
            # Australian EST.
            # assert_(t == '2012-01-20T22:05:32')

            # # what happens if you have too many frame bins
            # # you should get the same number of fr
            # a = PlatypusReduce(
            #     os.path.join(self.pth, 'PLP0011613.nx.hdf'),
            #     reflect=os.path.join(self.pth, 'PLP0011641.nx.hdf'),
            #     integrate=0, rebin_percent=2,
            #     eventmode=[0, 25200, 27000, 30000])
            # assert_equal(a.ydata.shape[0], 3)


class TestSpatzReduce:
    @pytest.mark.usefixtures("no_data_directory")
    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir, data_directory):
        self.pth = pjoin(data_directory, "reduce")

        self.cwd = os.getcwd()
        self.tmpdir = tmpdir.strpath
        os.chdir(self.tmpdir)
        return 0

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_smoke(self):
        # a quick smoke test to check that the reduction can occur
        # warnings filter for pixel size
        a, fname = reduce_stitch(
            [660, 661],
            [658, 659],
            data_folder=self.pth,
            prefix="SPZ",
            reduction_options={"rebin_percent": 2},
        )
        a.save("test1.dat")
        assert os.path.isfile("./test1.dat")

        # reduce_stitch should take a list of ReductionOptions dict,
        # separate dicts are used for different angles
        opts = ReductionOptions()
        opts["rebin_percent"] = 2

        a2, fname = reduce_stitch(
            [660, 661],
            [658, 659],
            data_folder=self.pth,
            prefix="SPZ",
            reduction_options=[opts] * 2,
        )
        a2.save("test2.dat")
        assert os.path.isfile("./test2.dat")
        assert_allclose(a.y, a2.y)

    def test_reduction_method(self):
        # a quick smoke test to check that the reduction can occur
        a = SpatzReduce("SPZ0000658.nx.hdf", data_folder=self.pth)

        # try reduction with the reduce method
        a.reduce(
            "SPZ0000660.nx.hdf",
            data_folder=self.pth,
            rebin_percent=4,
        )

        # try reduction with the __call__ method
        a(
            "SPZ0000660.nx.hdf",
            data_folder=self.pth,
            rebin_percent=4,
        )

        # this should also have saved a couple of files in the current
        # directory
        assert os.path.isfile("./SPZ0000660_0.dat")
        assert os.path.isfile("./SPZ0000660_0.xml")

        # try writing offspecular data
        a.write_offspecular("offspec.xml", 0)


class TestPolarisedReduce:
    @pytest.mark.usefixtures("no_data_directory")
    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir, data_directory):
        self.pth = pjoin(data_directory, "reduce", "PNR_files")

        self.cwd = os.getcwd()
        self.tmpdir = tmpdir.strpath
        os.chdir(self.tmpdir)
        return 0

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_polarised_reduction_method_4sc(self):
        # a quick smoke test to check that the reduction can occur
        # warnings filter for pixel size
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            spinset_rb = SpinSet(
                down_down=pjoin(self.pth, "PLP0012785.nx.hdf"),
                up_up=pjoin(self.pth, "PLP0012787.nx.hdf"),
                up_down=pjoin(self.pth, "PLP0012786.nx.hdf"),
                down_up=pjoin(self.pth, "PLP0012788.nx.hdf"),
            )
            spinset_db = SpinSet(
                down_down=pjoin(self.pth, "PLP0012793.nx.hdf"),
                up_up=pjoin(self.pth, "PLP0012795.nx.hdf"),
                up_down=pjoin(self.pth, "PLP0012794.nx.hdf"),
                down_up=pjoin(self.pth, "PLP0012796.nx.hdf"),
            )
            a = PolarisedReduce(spinset_db)

            # try reduction with the reduce method
            a.reduce(
                spinset_rb,
                data_folder=self.pth,
                lo_wavelength=2.5,
                hi_wavelength=12.5,
                rebin_percent=4,
            )
            # try reduction with the __call__ method
            a(
                spinset_rb,
                data_folder=self.pth,
                lo_wavelength=2.5,
                hi_wavelength=12.5,
                rebin_percent=4,
            )

            # this should also have saved a couple of files in the current
            # directory
            assert os.path.isfile("./PLP0012785_0_PolCorr.dat")
            assert os.path.isfile("./PLP0012786_0_PolCorr.dat")
            assert os.path.isfile("./PLP0012787_0_PolCorr.dat")
            assert os.path.isfile("./PLP0012788_0_PolCorr.dat")

            # can we read the file
            dd = ReflectDataset("./PLP0012785_0_PolCorr.dat")
            uu = ReflectDataset("./PLP0012787_0_PolCorr.dat")
            ud = ReflectDataset("./PLP0012786_0_PolCorr.dat")
            du = ReflectDataset("./PLP0012788_0_PolCorr.dat")

            # check if the written data is the same as what is in the reducers
            assert_equal(dd.x, list(reversed(a.reducers["dd"].x[0])))
            assert_equal(du.x, list(reversed(a.reducers["du"].x[0])))
            assert_equal(ud.x, list(reversed(a.reducers["ud"].x[0])))
            assert_equal(uu.x, list(reversed(a.reducers["uu"].x[0])))

            assert_equal(dd.y, list(reversed(a.reducers["dd"].y_corr[0])))
            assert_equal(du.y, list(reversed(a.reducers["du"].y_corr[0])))
            assert_equal(ud.y, list(reversed(a.reducers["ud"].y_corr[0])))
            assert_equal(uu.y, list(reversed(a.reducers["uu"].y_corr[0])))

            assert_equal(
                a.reducers["dd"].direct_beam.m_spec,
                a.reducers["du"].direct_beam.m_spec,
            )
            assert_equal(
                a.reducers["uu"].direct_beam.m_spec,
                a.reducers["ud"].direct_beam.m_spec,
            )
        # Check that the shapes of all the spectra that need to be the
        # same are the same.
        for sc in a._reduced_successfully:
            reducer = a.reducers[sc]
            assert (
                reducer.reflected_beam.m_spec_polcorr.shape
                == reducer.reflected_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec_polcorr.shape
                == reducer.direct_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec_polcorr.shape
                == reducer.reflected_beam.m_spec_polcorr.shape
            )
            assert (
                reducer.reflected_beam.m_spec.shape
                == reducer.reflected_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec.shape
                == reducer.direct_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec.shape
                == reducer.reflected_beam.m_spec.shape
            )

            # Make sure there is a difference between m_spec and
            # m_spec_polcorr
            with pytest.raises(AssertionError):
                assert_equal(
                    reducer.reflected_beam.m_spec,
                    reducer.reflected_beam.m_spec_polcorr,
                )

    def test_polarised_reduction_method_3sc(self):
        # a quick smoke test to check that the reduction can occur
        # warnings filter for pixel size
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            spinset_rb = SpinSet(
                down_down=pjoin(self.pth, "PLP0012785.nx.hdf"),
                up_up=pjoin(self.pth, "PLP0012787.nx.hdf"),
                up_down=pjoin(self.pth, "PLP0012786.nx.hdf"),
            )
            spinset_db = SpinSet(
                down_down=pjoin(self.pth, "PLP0012793.nx.hdf"),
                up_up=pjoin(self.pth, "PLP0012795.nx.hdf"),
                up_down=pjoin(self.pth, "PLP0012794.nx.hdf"),
            )
            a = PolarisedReduce(spinset_db)

            # try reduction with the reduce method
            a.reduce(
                spinset_rb,
                data_folder=self.pth,
                lo_wavelength=2.5,
                hi_wavelength=12.5,
                rebin_percent=4,
            )

            # try reduction with the __call__ method
            a(
                spinset_rb,
                data_folder=self.pth,
                lo_wavelength=2.5,
                hi_wavelength=12.5,
                rebin_percent=4,
            )

            # this should also have saved a couple of files in the current
            # directory
            assert os.path.isfile("./PLP0012785_0_PolCorr.dat")
            assert os.path.isfile("./PLP0012786_0_PolCorr.dat")
            assert os.path.isfile("./PLP0012787_0_PolCorr.dat")

            # can we read the file
            dd = ReflectDataset("./PLP0012785_0_PolCorr.dat")
            uu = ReflectDataset("./PLP0012787_0_PolCorr.dat")
            ud = ReflectDataset("./PLP0012786_0_PolCorr.dat")

            # check if the written data is the same as what is in the reducers
            # Note: the order of the data is reversed in the reducers
            # compared to the .dat file
            assert_equal(dd.x, list(reversed(a.reducers["dd"].x[0])))
            assert_equal(ud.x, list(reversed(a.reducers["ud"].x[0])))
            assert_equal(uu.x, list(reversed(a.reducers["uu"].x[0])))

            assert_equal(dd.y, list(reversed(a.reducers["dd"].y_corr[0])))
            assert_equal(ud.y, list(reversed(a.reducers["ud"].y_corr[0])))
            assert_equal(uu.y, list(reversed(a.reducers["uu"].y_corr[0])))

            assert_equal(
                a.reducers["dd"].direct_beam.m_spec,
                a.reducers["du"].direct_beam.m_spec,
            )
            assert_equal(
                a.reducers["uu"].direct_beam.m_spec,
                a.reducers["ud"].direct_beam.m_spec,
            )
        # Check that the shapes of all the spectra that need to be the
        # same are the same.
        for sc in a._reduced_successfully:
            reducer = a.reducers[sc]
            assert (
                reducer.reflected_beam.m_spec_polcorr.shape
                == reducer.reflected_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec_polcorr.shape
                == reducer.direct_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec_polcorr.shape
                == reducer.reflected_beam.m_spec_polcorr.shape
            )
            assert (
                reducer.reflected_beam.m_spec.shape
                == reducer.reflected_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec.shape
                == reducer.direct_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec.shape
                == reducer.reflected_beam.m_spec.shape
            )

            # Make sure there is a difference between m_spec and
            # m_spec_polcorr
            with pytest.raises(AssertionError):
                assert_equal(
                    reducer.reflected_beam.m_spec,
                    reducer.reflected_beam.m_spec_polcorr,
                )

    def test_polarised_reduction_method_2sc(self):
        # a quick smoke test to check that the reduction can occur
        # warnings filter for pixel size
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            spinset_rb = SpinSet(
                down_down=pjoin(self.pth, "PLP0012785.nx.hdf"),
                up_up=pjoin(self.pth, "PLP0012787.nx.hdf"),
            )
            spinset_db = SpinSet(
                down_down=pjoin(self.pth, "PLP0012793.nx.hdf"),
                up_up=pjoin(self.pth, "PLP0012795.nx.hdf"),
            )
            a = PolarisedReduce(spinset_db)

            # try reduction with the reduce method
            a.reduce(
                spinset_rb,
                lo_wavelength=2.5,
                hi_wavelength=12.5,
                rebin_percent=3,
            )

            # try reduction with the __call__ method
            a(
                spinset_rb,
                lo_wavelength=2.5,
                hi_wavelength=12.5,
                rebin_percent=3,
            )

            # this should also have saved a couple of files in the current
            # directory
            assert os.path.isfile("./PLP0012785_0_PolCorr.dat")
            assert os.path.isfile("./PLP0012787_0_PolCorr.dat")

            # can we read the file
            dd = ReflectDataset("./PLP0012785_0_PolCorr.dat")
            uu = ReflectDataset("./PLP0012787_0_PolCorr.dat")

            # check if the written data is the same as what is in the reducers
            assert_equal(dd.x, list(reversed(a.reducers["dd"].x[0])))
            assert_equal(uu.x, list(reversed(a.reducers["uu"].x[0])))

            assert_equal(dd.y, list(reversed(a.reducers["dd"].y_corr[0])))
            assert_equal(uu.y, list(reversed(a.reducers["uu"].y_corr[0])))

            assert_equal(
                a.reducers["dd"].direct_beam.m_spec,
                a.reducers["du"].direct_beam.m_spec,
            )
            assert_equal(
                a.reducers["uu"].direct_beam.m_spec,
                a.reducers["ud"].direct_beam.m_spec,
            )
        # Check that the shapes of all the spectra that need to be the
        # same are the same.
        for sc in a._reduced_successfully:
            reducer = a.reducers[sc]
            assert (
                reducer.reflected_beam.m_spec_polcorr.shape
                == reducer.reflected_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec_polcorr.shape
                == reducer.direct_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec_polcorr.shape
                == reducer.reflected_beam.m_spec_polcorr.shape
            )
            assert (
                reducer.reflected_beam.m_spec.shape
                == reducer.reflected_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec.shape
                == reducer.direct_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec.shape
                == reducer.reflected_beam.m_spec.shape
            )

            # Make sure there is a difference between m_spec and
            # m_spec_polcorr
            with pytest.raises(AssertionError):
                assert_equal(
                    reducer.reflected_beam.m_spec,
                    reducer.reflected_beam.m_spec_polcorr,
                )

    def test_nsf_spin_channels(self):
        """
        Check that the efficiency-corrected R++ channel is assigned properly
        in the reducer by comparing where the reflectivity drops below a
        certain threshold
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            spinset_rb = SpinSet(
                down_down=pjoin(self.pth, "PLP0012785.nx.hdf"),
                up_up=pjoin(self.pth, "PLP0012787.nx.hdf"),
            )
            spinset_db = SpinSet(
                down_down=pjoin(self.pth, "PLP0012793.nx.hdf"),
                up_up=pjoin(self.pth, "PLP0012795.nx.hdf"),
            )
            a = PolarisedReduce(spinset_db)

            # Reduce and correct data
            a.reduce(
                spinset_rb,
                lo_wavelength=2.5,
                hi_wavelength=12.5,
                rebin_percent=3,
            )

            # Get Q position where the reflectivity drops to 0.5 for the
            # "uu" and "dd" datasets
            q_dd = a.reducers["dd"].x[0][
                abs(a.reducers["dd"].y - 0.5).argmin()
            ]
            q_uu = a.reducers["uu"].x[0][
                abs(a.reducers["uu"].y - 0.5).argmin()
            ]
            # Check x_uu is larger than x_dd. This is because the spin up
            # neutrons see a higher potential for the magnetically saturated
            # permalloy film. This checks that the "uu" and "dd" spin channels
            # didn't get mixed up during the reduction process
            assert q_uu > q_dd


class TestPolarisationEfficiency:
    @pytest.mark.usefixtures("no_data_directory")
    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir, data_directory):
        self.pth = pjoin(data_directory, "reduce", "PNR_files")

        self.cwd = os.getcwd()
        self.tmpdir = tmpdir.strpath
        os.chdir(self.tmpdir)
        return 0

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_smoke(self):
        wavelength_axis = calculate_wavelength_bins(2.5, 12.5, 3)

        peff = PolarisationEfficiency(wavelength_axis)

        assert peff.combined_efficiency_matrix.shape == tuple(
            [len(wavelength_axis), 4, 4]
        )

    def test_config_difference(self):
        wavelength_axis = calculate_wavelength_bins(2.5, 12.5, 3)

        p_PF = PolarisationEfficiency(wavelength_axis, config="PF")
        p_full = PolarisationEfficiency(wavelength_axis, config="full")

        with pytest.raises(AssertionError):
            assert_equal(
                p_PF.combined_efficiency_matrix,
                p_full.combined_efficiency_matrix,
            )

    def test_input(self):
        wavelength_axis = calculate_wavelength_bins(2.5, 12.5, 3).reshape(
            1, -1
        )

        with pytest.raises(ValueError):
            peff = PolarisationEfficiency(wavelength_axis)

            assert peff

    def test_manual_input_to_PolarisedReduce(self):
        """
        Test the manual input of pol_eff into a
        `refnx.reduce.PolarisedReduce` object and see if it is the same as
        an automatically created one
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            # Create SpinSets, PolarisedReduce object and ReductionOptions
            spinset_rb = SpinSet(
                down_down=pjoin(self.pth, "PLP0012785.nx.hdf"),
                up_up=pjoin(self.pth, "PLP0012787.nx.hdf"),
            )
            spinset_db = SpinSet(
                down_down=pjoin(self.pth, "PLP0012793.nx.hdf"),
                up_up=pjoin(self.pth, "PLP0012795.nx.hdf"),
            )
            a = PolarisedReduce(spinset_db)

            rdo = ReductionOptions(
                lo_wavelength=2.5,
                hi_wavelength=12.5,
                rebin_percent=3,
            )

            # Reduce and correct reflected beams
            a.reduce(spinset_rb, **rdo)

            # Create another polarised reducer to compare
            b = PolarisedReduce(spinset_db)
            # Create PolarisationEfficiency object and reduce
            pol_eff = PolarisationEfficiency(
                a.reducers["dd"].direct_beam.m_lambda[0],
                config="full",
            )
            b.reduce(spinset_rb, pol_eff=pol_eff, **rdo)

            for sc in a._reduced_successfully:
                assert_equal(a.reducers[sc].x, b.reducers[sc].x)
                assert_equal(a.reducers[sc].y, b.reducers[sc].y)
                assert_equal(a.reducers[sc].y_err, b.reducers[sc].y_err)
                assert_equal(a.reducers[sc].y_corr, b.reducers[sc].y_corr)



import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_equal,
    assert_,
    assert_allclose,
)
from scipy.stats import norm
from refnx.reflect import Erf, Exponential, Step, Linear, Tanh, Sinusoidal


class TestStructure:
    def setup_method(self):
        self.x = np.linspace(-5, 5, 1001)

    def test_erf(self):
        i = Erf()
        profile = i(self.x, scale=1.1, loc=-1.0)

        assert_equal(profile, norm.cdf(self.x, scale=1.1, loc=-1.0))

    def test_exp(self):
        i = Exponential()
        i(self.x, scale=1.1, loc=-1.0)

    def test_linear(self):
        i = Linear()
        i(self.x, scale=1.1, loc=-1.0)

    def test_step(self):
        i = Step()
        i(self.x, scale=1.1, loc=-1.0)

    def test_sin(self):
        i = Sinusoidal()
        i(self.x, scale=1.1, loc=-1.0)

    def test_Tanh(self):
        i = Tanh()
        i(self.x, scale=1.1, loc=-1.0)

    def test_repr(self):
        cls = [Erf, Exponential, Step, Linear, Tanh, Sinusoidal]

        for c in cls:
            o = c()
            p = eval(repr(o))
            assert isinstance(p, c)



import string
from copy import deepcopy
import os.path
from time import gmtime, strftime
from multiprocessing import Queue
from threading import Thread
import time

import numpy as np
import pandas as pd
import h5py

from refnx.reduce.platypusnexus import (
    PlatypusNexus,
    ReflectNexus,
    number_datafile,
    basename_datafile,
    SpatzNexus,
    ReductionOptions,
    calculate_wavelength_bins,
)
from refnx.util import ErrorProp as EP
import refnx.util.general as general
from refnx.reduce.parabolic_motion import (
    parabola_line_intersection_point,
    find_trajectory,
)
from refnx.dataset import ReflectDataset
from refnx._lib import possibly_open_file


_template_ref_xml = """<?xml version="1.0"?>
<REFroot xmlns="">
<REFentry time="$time">
<Title>$title</Title>
<User>$user</User>
<REFsample>
<ID>$sample</ID>
</REFsample>
<REFdata axes="Qz:Qx" rank="2" type="POINT" \
spin="UNPOLARISED" dim="$_numpointsz:$_numpointsy">
<Run filename="$_rnumber" preset="" size="">
</Run>
<R uncertainty="dR">$_r</R>
<Qz uncertainty="dQz" units="1/A">$_qz</Qz>
<dR type="SD">$_dr</dR>
<Qx type="_FWHM" units="1/A">$_qx</Qx>
</REFdata>
</REFentry>
</REFroot>"""


class ReflectReduce:
    def __init__(self, direct, prefix, data_folder=None):

        self.data_folder = os.path.curdir
        if data_folder is not None:
            self.data_folder = data_folder

        if prefix == "PLP":
            self.reflect_klass = PlatypusNexus
        elif prefix == "SPZ":
            self.reflect_klass = SpatzNexus
        else:
            raise ValueError(
                "Instrument prefix not known. Must be one of" " ['PLP']"
            )

        if isinstance(direct, ReflectNexus):
            self.direct_beam = direct
        elif type(direct) is str:
            direct = os.path.join(self.data_folder, direct)
            self.direct_beam = self.reflect_klass(direct)
        else:
            self.direct_beam = self.reflect_klass(direct)

        self.prefix = prefix

    def __call__(self, reflect, scale=1.0, save=True, **reduction_options):
        return self.reduce(
            reflect, scale=scale, save=save, **reduction_options
        )

    def reduce(self, reflect, scale=1.0, save=True, **reduction_options):
        """
        Reduction of a single dataset.

        The reduction uses the direct beam specified during construction of
        this object. This method reduces all the spectra present in the
        reflected beam file (see platypusnexus.PlatypusNexus.process for
        eventmode specification and other related options), but aggregates
        all data in the direct beam spectrum.

        Parameters
        ----------
        reflect : string, hdf5 file-handle or PlatypusNexus object
            A string containing the path to the specularly reflected hdf5 file,
            the hdf5 file itself, or a PlatypusNexus object.
        scale : float, optional
            Divide all the reflectivity values by this number.
        save : bool, optional
            If `True` then the reduced dataset is saved to the current
            directory, with a name os.path.basename(reflect)
        reduction_options : dict, optional
            Options passed directly to `refnx.reduce.PlatypusNexus.process`,
            for processing of individual spectra. Look at that method docstring
            for specification of options.

        Returns
        -------
        datasets, reduction : tuple

        datasets : sequence of ReflectDataset

        reduction : dict
            Contains the following entries:

            - 'x' : np.ndarray
                Q values, shape (N, T).
            - 'x_err' : np.ndarray
                Uncertainty in Q values (FWHM), shape (N, T).
            - 'y' : np.ndarray
                Specular Reflectivity, shape (N, T)
            - 'y_err' : np.ndarray
                Uncertainty in specular reflectivity (SD), shape (N, T)
            - 'omega' : np.ndarray
                Angle of incidence, shape (N, T)
            - 'm_lambda' : np.ndarray
                Wavelength, shape (N, T)
            - 'm_ref' : np.ndarray
                Offspecular reflectivity map, shape (N, T, Y)
            - 'm_ref_err' : np.ndarray
                uncertainty in offspecular reflectivity, shape (N, T, Y)
            - 'm_qz' : np.ndarray
                Qz for offspecular map, shape (N, T, Y)
            - 'm_qx' : np.ndarray
                Qx for offspecular map, shape (N, T, Y)
            - 'n_spectra' : int
                number of reflectivity spectra
            - 'datafile_number' : int
                run number for the reflected beam
            - 'fname' : list
                the saved filenames
            - 'direct_beam' : ReflectNexus
                Direct Beam object
            - 'reflected_beam' : ReflectNexus
                Reflected Beam object

        N corresponds to the number of spectra
        T corresponds to the number of Q (wavelength) bins
        Y corresponds to the number of y pixels on the detector.

        Notes
        -----
        All the values returned from this method are also contained as instance
        attributes for this object.

        Examples
        --------

        >>> from refnx.reduce import PlatypusReduce
        >>> # set up with a direct beam
        >>> reducer = PlatypusReduce('PLP0000711.nx.hdf')
        >>> datasets, reduction = reducer.reduce('PLP0000708.nx.hdf',
        ...                                      rebin_percent=3.)

        """
        reflect_keywords = reduction_options.copy()
        direct_keywords = reduction_options.copy()

        # get the direct beam spectrum
        direct_keywords["direct"] = True
        direct_keywords["integrate"] = -1
        if (
            "peak_pos" in direct_keywords
            and hasattr(direct_keywords["peak_pos"], "len")
            and len(direct_keywords["peak_pos"]) == 2
        ):
            # don't use a user specified peak_pos for direct and reflected
            # beams, only for the reflected beam. Leave the computer to find
            # the direct beam pos. Alternatively one can use the manual beam
            # finder
            direct_keywords.pop("peak_pos")

        if "eventmode" in direct_keywords:
            direct_keywords.pop("eventmode")

        if "event_filter" in direct_keywords:
            direct_keywords.pop("event_filter")

        self.direct_beam.process(**direct_keywords)

        # get the reflected beam spectrum
        reflect_keywords["direct"] = False
        if isinstance(reflect, ReflectNexus):
            self.reflected_beam = reflect
        elif type(reflect) is str:
            reflect = os.path.join(self.data_folder, reflect)
            self.reflected_beam = self.reflect_klass(reflect)
        else:
            self.reflected_beam = self.reflect_klass(reflect)

        # Got to use the same wavelength bins as the direct spectrum.
        # done this way around to save processing direct beam over and over
        reflect_keywords["wavelength_bins"] = self.direct_beam.m_lambda_hist[0]

        self.reflected_beam.process(**reflect_keywords)

        self.save = save
        self._reduce_single_angle(scale)

        """
        --Specular Reflectivity--
        Use the (constant wavelength) spectra that have already been integrated
        over 2theta (in processnexus) to calculate the specular reflectivity.
        Beware: this is because m_topandtail has already been divided through
        by monitor counts and error propagated (at the end of processnexus).
        Thus, the 2theta pixels are correlated to some degree. If we use the 2D
        plot to calculate reflectivity
        (sum {Iref_{2theta, lambda}}/I_direct_{lambda}) then the error bars in
        the reflectivity turn out much larger than they should be.
        """
        ydata, ydata_sd = EP.EPdiv(
            self.reflected_beam.m_spec,
            self.reflected_beam.m_spec_sd,
            self.direct_beam.m_spec,
            self.direct_beam.m_spec_sd,
        )

        # calculate the 1D Qz values.
        xdata = general.q(self.omega_corrected, self.reflected_beam.m_lambda)
        xdata_sd = (
            self.reflected_beam.m_lambda_fwhm / self.reflected_beam.m_lambda
        ) ** 2
        xdata_sd += (
            self.reflected_beam.domega[:, np.newaxis] / self.omega_corrected
        ) ** 2
        xdata_sd = np.sqrt(xdata_sd) * xdata

        """
        ---Offspecular reflectivity---
        normalise the counts in the reflected beam by the direct beam
        spectrum this gives a reflectivity. Also propagate the errors,
        leaving the fractional variance (dr/r)^2.
        --Note-- that adjacent y-pixels (same wavelength) are correlated in
        this treatment, so you can't just sum over them.
        i.e. (c_0 / d) + ... + c_n / d) != (c_0 + ... + c_n) / d
        """
        m_ref, m_ref_sd = EP.EPdiv(
            self.reflected_beam.m_topandtail,
            self.reflected_beam.m_topandtail_sd,
            self.direct_beam.m_spec[:, :, np.newaxis],
            self.direct_beam.m_spec_sd[:, :, np.newaxis],
        )

        # you may have had divide by zero's.
        m_ref = np.where(np.isinf(m_ref), 0, m_ref)
        m_ref_sd = np.where(np.isinf(m_ref_sd), 0, m_ref_sd)

        # calculate the Q values for the detector pixels.  Each pixel has
        # different 2theta and different wavelength, ASSUME that they have the
        # same angle of incidence
        qx, qy, qz = general.q2(
            self.omega_corrected[:, :, np.newaxis],
            self.m_twotheta,
            0,
            self.reflected_beam.m_lambda[:, :, np.newaxis],
        )

        reduction = {}
        reduction["x"] = self.x = xdata
        reduction["x_err"] = self.x_err = xdata_sd
        reduction["y"] = self.y = ydata / scale
        reduction["y_err"] = self.y_err = ydata_sd / scale
        reduction["m_ref"] = self.m_ref = m_ref
        reduction["m_ref_err"] = self.m_ref_err = m_ref_sd
        reduction["qz"] = self.m_qz = qz
        reduction["qx"] = self.m_qx = qx
        reduction["omega"] = self.omega_corrected
        reduction["m_lambda"] = self.reflected_beam.m_lambda
        reduction["nspectra"] = self.n_spectra
        reduction["start_time"] = self.reflected_beam.start_time
        reduction[
            "datafile_number"
        ] = self.datafile_number = self.reflected_beam.datafile_number
        reduction["direct_beam"] = self.direct_beam
        reduction["reflected_beam"] = self.reflected_beam

        fnames = []
        datasets = []
        datafilename = self.reflected_beam.datafilename
        datafilename = os.path.basename(datafilename.split(".nx.hdf")[0])

        header = self._create_metadata_header()

        for i in range(self.n_spectra):
            data_tup = self.data(scanpoint=i)
            datasets.append(ReflectDataset(data_tup))

        if self.save:
            for i, dataset in enumerate(datasets):
                fname = f"{datafilename}_{i}.dat"
                fnames.append(fname)
                with open(fname, "wb") as f:
                    dataset.save(f, header=header)

                fname = f"{datafilename}_{i}.xml"
                with open(fname, "wb") as f:
                    dataset.save_xml(f, start_time=reduction["start_time"][i])
        reduction["fname"] = fnames
        return datasets, deepcopy(reduction)

    def data(self, scanpoint=0):
        """
        The specular reflectivity

        Parameters
        ----------
        scanpoint: int
            Find a particular specular reflectivity image. scanpoints upto
            `self.n_spectra - 1` can be specified.

        Returns
        -------
        (Q, R, dR, dQ): np.ndarray tuple
            dR is standard deviation, dQ is FWHM
        """
        return (
            self.x[scanpoint],
            self.y[scanpoint],
            self.y_err[scanpoint],
            self.x_err[scanpoint],
        )

    def data2d(self, scanpoint=0):
        """
        The offspecular data

        Parameters
        ----------
        scanpoint: int
            Find a particular offspecular image. scanpoints upto
            self.n_spectra - 1 can be specified.

        Returns
        -------
        (Qz, Qx, R, dR): np.ndarrays
        """

        return (
            self.m_qz[scanpoint],
            self.m_qx[scanpoint],
            self.m_ref[scanpoint],
            self.m_ref_err[scanpoint],
        )

    def scale(self, scale):
        """
        Divides the reflectivity values by this scale factor

        Parameters
        ----------
        scale: float
            Divides the reflectivity values by a constant amount
        """
        self.m_ref /= scale
        self.m_ref_err /= scale
        self.y /= scale
        self.y_err /= scale

    def write_offspecular(self, f, scanpoint=0):
        d = dict()
        d["time"] = strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())
        d["_rnumber"] = self.reflected_beam.datafile_number
        d["_numpointsz"] = np.size(self.m_ref, 1)
        d["_numpointsy"] = np.size(self.m_ref, 2)

        s = string.Template(_template_ref_xml)

        # filename = 'off_PLP{:07d}_{:d}.xml'.format(self._rnumber, index)
        d["_r"] = repr(self.m_ref[scanpoint].tolist()).strip(",[]")
        d["_qz"] = repr(self.m_qz[scanpoint].tolist()).strip(",[]")
        d["_dr"] = repr(self.m_ref_err[scanpoint].tolist()).strip(",[]")
        d["_qx"] = repr(self.m_qx[scanpoint].tolist()).strip(",[]")

        thefile = s.safe_substitute(d)

        with possibly_open_file(f, "wb") as g:
            if "b" in g.mode:
                thefile = thefile.encode("utf-8")

            g.write(thefile)
            g.truncate()

    def _create_metadata_header(self):
        header = []
        header.append(
            f"reflected_beam_number: {self.reflected_beam.datafilename}"
        )
        header.append(f"direct_run_number: {self.direct_beam.datafilename}")
        header.append(f"samplename: {self.reflected_beam.cat.sample_name}")

        ps_dct = self.reflected_beam.processed_spectrum
        header.append(f"beam_centre: {ps_dct['m_beampos']}")
        header.append(f"beam_sd: {ps_dct['m_beampos_sd']}")
        header.append(f"lopx: {ps_dct['hipx']}")
        header.append(f"hipx: {ps_dct['lopx']}")
        rdo = ps_dct["reduction_options"]
        header.append(f"rebin_percent: {rdo['rebin_percent']}")
        header.append(f"background: {rdo['background']}")
        header.append(f"lo_wavelength: {rdo['lo_wavelength']}")
        header.append(f"hi_wavelength: {rdo['hi_wavelength']}")

        header.append(
            "Warning: the format of this header may change at any"
            " time. Do not rely on it staying constant"
        )
        header.append("Q (1/A), R, dR (sigma), dQ (1/A, FWHM)")
        return "\n".join(header)


class PlatypusReduce(ReflectReduce):
    """
    Reduces Platypus reflectometer data to give the specular reflectivity.
    Offspecular data maps are also produced.

    Parameters
    ----------
    direct : string, hdf5 file-handle or PlatypusNexus object
        A string containing the path to the direct beam hdf5 file,
        the hdf5 file itself, or a PlatypusNexus object.
    data_folder : str, optional
        Folder where the datafiles are stored.

    Examples
    --------

    >>> from refnx.reduce import PlatypusReduce
    >>> reducer = PlatypusReduce('PLP0000711.nx.hdf')
    >>> datasets, reduced = reducer.reduce('PLP0000711.nx.hdf',
    ...                                    rebin_percent=2)

    """

    def __init__(self, direct, data_folder=None, **kwds):

        super().__init__(direct, "PLP", data_folder=data_folder)

    def _reduce_single_angle(self, scale=1):
        """
        Reduce a single angle.
        """
        n_spectra = self.reflected_beam.n_spectra
        n_tpixels = np.size(self.reflected_beam.m_topandtail, 1)
        n_ypixels = np.size(self.reflected_beam.m_topandtail, 2)

        # calculate omega and two_theta depending on the mode.
        mode = self.reflected_beam.mode

        # we'll need the wavelengths to calculate gravity effects.
        wavelengths = self.reflected_beam.m_lambda
        m_twotheta = np.zeros((n_spectra, n_tpixels, n_ypixels))

        detector_z_difference = (
            self.reflected_beam.detector_z - self.direct_beam.detector_z
        )

        beampos_z_difference = (
            self.reflected_beam.m_beampos - self.direct_beam.m_beampos
        )

        Y_PIXEL_SPACING = self.reflected_beam.cat.qz_pixel_size[0]

        total_z_deflection = (
            detector_z_difference + beampos_z_difference * Y_PIXEL_SPACING
        )

        if mode in ["FOC", "POL", "POLANAL", "MT"]:
            # omega_nom.shape = (N, )
            omega_nom = np.degrees(
                np.arctan(total_z_deflection / self.reflected_beam.detector_y)
                / 2.0
            )

            """
            Wavelength specific angle of incidence correction
            This involves:
            1) working out the trajectory of the neutrons through the
            collimation system.
            2) where those neutrons intersect the sample.
            3) working out the elevation of the neutrons when they hit the
            sample.
            4) correcting the angle of incidence.
            """
            speeds = general.wavelength_velocity(wavelengths)
            collimation_distance = self.reflected_beam.cat.collimation_distance
            s2_sample_distance = (
                self.reflected_beam.cat.sample_distance
                - self.reflected_beam.cat.slit2_distance
            )

            # work out the trajectories of the neutrons for them to pass
            # through the collimation system.
            trajectories = find_trajectory(
                collimation_distance / 1000.0, 0, speeds
            )

            # work out where the beam hits the sample
            res = parabola_line_intersection_point(
                s2_sample_distance / 1000,
                0,
                trajectories,
                speeds,
                omega_nom[:, np.newaxis],
            )
            intersect_x, intersect_y, x_prime, elevation = res

            # correct the angle of incidence with a wavelength dependent
            # elevation.
            omega_corrected = omega_nom[:, np.newaxis] - elevation

            m_twotheta += np.arange(n_ypixels * 1.0)[np.newaxis, np.newaxis, :]
            m_twotheta -= self.direct_beam.m_beampos[:, np.newaxis, np.newaxis]
            m_twotheta *= Y_PIXEL_SPACING
            m_twotheta += detector_z_difference
            m_twotheta /= self.reflected_beam.detector_y[
                :, np.newaxis, np.newaxis
            ]
            m_twotheta = np.arctan(m_twotheta)
            m_twotheta = np.degrees(m_twotheta)

            # you may be reflecting upside down, reverse the sign.
            upside_down = np.sign(omega_corrected[:, 0])
            m_twotheta *= upside_down[:, np.newaxis, np.newaxis]
            omega_corrected *= upside_down[:, np.newaxis]

        elif mode in ["SB", "DB"]:
            # the angle of incidence is half the two theta of the reflected
            # beam
            omega = (
                np.arctan(total_z_deflection / self.reflected_beam.detector_y)
                / 2.0
            )

            # work out two theta for each of the detector pixels
            m_twotheta += np.arange(n_ypixels * 1.0)[np.newaxis, np.newaxis, :]
            m_twotheta -= self.direct_beam.m_beampos[:, np.newaxis, np.newaxis]
            m_twotheta += detector_z_difference
            m_twotheta -= self.reflected_beam.detector_y[
                :, np.newaxis, np.newaxis
            ] * np.tan(omega[:, np.newaxis, np.newaxis])

            m_twotheta /= self.reflected_beam.detector_y[
                :, np.newaxis, np.newaxis
            ]
            m_twotheta = np.arctan(m_twotheta)
            m_twotheta += omega[:, np.newaxis, np.newaxis]

            # still in radians at this point
            # add an extra dimension, because omega_corrected needs to be the
            # angle of incidence for each wavelength. I.e. should be
            # broadcastable to (N, T)
            omega_corrected = np.degrees(omega)[:, np.newaxis]
            m_twotheta = np.degrees(m_twotheta)

        self.omega_corrected = omega_corrected
        self.m_twotheta = m_twotheta
        self.n_spectra = n_spectra


class PolarisationEfficiency:
    """
    Describes the polarisation efficiency of a neutron scattering system
    with the option of having a polariser, flipper-1, flipper-2, and
    analyser in the system.

    The dimensions of processed spectra are (n_spectra, TOF). Since
    we expect that our efficiencies are purely wavelength dependent
    and not dependent upon the spectra within a Nexus file, we neglect the
    n_spectra axis when dealing with polarisation efficiencies.

    Parameters
    ----------
    wavelength_axis :   numpy.array (T,)
                    Array of wavelength bin centres to initialise the length
                    of the (T, 4, 4) efficiency matrices.

    config          :   {"full", "PF"}
                    Indication of polariser/analyser configuration. If
                    "full" is used, all polarising and flipping elements are
                    taken into account. If "PF" is used, only the polariser
                    and flipper are taken into account.
    """

    def __init__(self, wavelength_axis, config="full"):
        # Define sizes matrices to be (T,4,4) where T is the number of
        # wavelength bins
        if np.ndim(wavelength_axis) != 1:
            raise ValueError(
                "Number of dimensions of wavelength axis must be 1"
            )
        self.wl = wavelength_axis
        self.pol_eff = np.empty_like(self.wl)
        self.ana_eff = np.empty_like(self.wl)
        self.flipper1_eff = np.empty_like(self.wl)
        self.flipper2_eff = np.empty_like(self.wl)
        self.polariser_matrix = np.zeros((len(self.wl), 4, 4))
        self.analyser_matrix = np.zeros_like(self.polariser_matrix)
        self.flipper1_matrix = np.zeros_like(self.polariser_matrix)
        self.flipper2_matrix = np.zeros_like(self.polariser_matrix)
        self.combined_efficiency_matrix = np.empty_like(self.polariser_matrix)

        # Initialise standard values
        self.standard_efficiencies(config=config)

    def standard_efficiencies(self, config):
        """
        Define PLATYPUS polarisation efficiency matrices as described in
        the invited article in Rev. Sci. Instr. 83, 081301 (2012)
        `Polarization "Down Under": The polarized time-of-flight neutron
        reflectometer PLATYPUS' (https://doi.org/10.1063/1.4738579).

        In this formulation, the relationship between raw spectra
        from each spin channel and the efficiency-corrected polarised
        reflectivity is shown by the matrix equation

        I = F1 * F2 * P * A * R

        where I and R are the (TOF, 4, 1) raw spectra and corrected reflectivity, &
        F1, F2, P, and A are the (TOF, 4, 4) efficiency matrices from the RF
        flippers, polariser and analyser.

        This includes coefficients for the function `f(x) = a - b * c ** x`

        Parameters
        ----------
        config      :   {"full", "PF"}
        """
        # Define polariser efficiency as function of wavelength.
        p1a = 0.993
        p1b = 0.57
        p1c = 0.47
        self.pol_eff = p1a - p1b * p1c ** (self.wl)

        # Define analyser efficiency as function of wavelength
        p2a = 0.993
        p2b = 0.57
        p2c = 0.51
        self.ana_eff = p2a - p2b * p2c ** (self.wl)

        # Define flipper1 and flipper2 efficiencies as function of wavelength
        # These are set with a constant value as these are essentially
        # wavelength independent for PLATYPUS
        self.flipper1_eff = np.full(len(self.wl), 0.003)
        self.flipper2_eff = np.full(len(self.wl), 0.003)

        # Convert efficiencies to the form where:
        # P = 0 implies total spin polarisation in the down direction
        # P = 1/2 implies zero net spin polarisation
        # P = 1 implies total spin polarisation in the up direction
        P1 = (1 + self.pol_eff) / 2
        P2 = (1 + self.ana_eff) / 2

        F1 = self.flipper1_eff
        F2 = self.flipper2_eff

        # Check analyser position. If out of the beam, assume analyser and
        # flipper2 efficiency is perfect.
        if config == "PF":
            F2 = np.full(len(self.wl), 0.000)
            P2 = np.full(len(self.wl), 0.000)

        # Fill a (T, 4, 4) matrix for the polariser, analyser, flipper1,
        # and flipper2 efficiencies for each wavelength bin. Then
        # multiply them together for the combined efficiency

        # Create an array of zeros and ones the same length as the wavelength-dependent
        # polarisation efficiency array P1 to use in the vectorised
        # construction of the efficiency matrix
        z = np.zeros_like(P1)
        one = np.ones_like(P1)

        # Shape of P1 is (T,)
        # Polariser matrix shape is (4, 4, T). Transpose dimensions to be (T, 4, 4)
        self.polariser_matrix = [
            [(1 - P1), z, P1, z],
            [z, (1 - P1), z, P1],
            [P1, z, (1 - P1), z],
            [z, P1, z, (1 - P1)],
        ]
        self.polariser_matrix = np.transpose(
            self.polariser_matrix, axes=(2, 0, 1)
        )
        # Shape of P2 is (T,)
        # Analyser matrix shape is (4, 4, T). Transpose dimensions to be (T, 4, 4)
        self.analyser_matrix = [
            [(1 - P2), P2, z, z],
            [P2, (1 - P2), z, z],
            [z, z, (1 - P2), P2],
            [z, z, P2, (1 - P2)],
        ]
        self.analyser_matrix = np.transpose(
            self.analyser_matrix, axes=(2, 0, 1)
        )
        # Shape of F1 is (T,)
        # Flipper 1 matrix shape is (4, 4, T). Transpose dimensions to be (T, 4, 4)
        self.flipper1_matrix = [
            [one, z, z, z],
            [z, one, z, z],
            [F1, z, (1 - F1), z],
            [z, F1, z, (1 - F1)],
        ]
        self.flipper1_matrix = np.transpose(
            self.flipper1_matrix, axes=(2, 0, 1)
        )
        # Shape of F2 is (T,)
        # Flipper 2 matrix shape is (4, 4, T). Transpose dimensions to be (T, 4, 4)
        self.flipper2_matrix = [
            [one, z, z, z],
            [F2, (1 - F2), z, z],
            [z, z, one, z],
            [z, z, F2, (1 - F2)],
        ]
        self.flipper2_matrix = np.transpose(
            self.flipper2_matrix, axes=(2, 0, 1)
        )

        # Broadcasted matrix multiplication of efficiency matrices
        # Shape is (T, 4, 4). This is to be applied to a (N, T, 4, 1) array
        # of the measured spin channel intensities to produce a (N, T, 4, 1)
        # array of the efficiency-corrected spectra for each spin channel
        self.combined_efficiency_matrix = (
            self.flipper1_matrix
            @ self.flipper2_matrix
            @ self.polariser_matrix
            @ self.analyser_matrix
        )

    def custom_efficiencies(self, config):
        """
        Define custom efficiency function for polariser, analyser,
        and flippers to reduce data. Recommended only for advanced users.
        """
        raise NotImplementedError


class PolarisedReduce:
    """
    Reduces a direct beam and reflected beam spinset to produce
    a polarised neutron reflectivity curve that is corrected
    for polarisation efficiency.

    Parameters
    ----------
    spin_set_direct :   refnx.reduce.SpinSet
        Direct beam runs from a PNR experiment.

    Attributes
    ----------
    spin_set_direct :   refnx.reduce.SpinSet
        Direct beams from PNR experiment
    reducers        :   dict
            Dictionary of each measured spin channel
                "dd"    :   refnx.reduce.PlatypusNexus (R--)
                "du"    :   refnx.reduce.PlatypusNexus or None (R-+)
                "ud"    :   refnx.reduce.PlatypusNexus or None (R+-)
                "uu"    :   refnx.reduce.PlatypusNexus (R++)

    Examples
    --------
    >>> from refnx.reduce import SpinSet, PolarisedReduce
    >>> direct_beams = SpinSet(
    ...     down_down = 'PLP0012793.nx.hdf',
    ...     up_up = 'PLP0012795.nx.hdf',
    ...     up_down = 'PLP0012794.nx.hdf',
    ...     down_up = 'PLP0012796.nx.hdf'
    ... )
    >>> refl_beams = SpinSet(
    ...     down_down = 'PLP0012785.nx.hdf',
    ...     up_up = 'PLP0012787.nx.hdf',
    ...     up_down = 'PLP0012786.nx.hdf',
    ...     down_up = 'PLP0012788.nx.hdf'
    ... )
    >>> reducer = PolarisedReduce(direct_beams)
    >>> datasets, reduced = reducer.reduce(refl_beams)
    """

    def __init__(self, spin_set_direct):
        self.spin_set_direct = spin_set_direct
        self.reducers = {}
        # Note: order of dd, du, ud, uu matters here since we iterate
        # over these later on
        for sc in ["dd", "du"]:
            self.reducers[sc] = PlatypusReduce(spin_set_direct.dd)
        for sc in ["ud", "uu"]:
            self.reducers[sc] = PlatypusReduce(spin_set_direct.uu)

    def __call__(self, spin_set_reflect, pol_eff=None, **reduction_options):
        return self.reduce(spin_set_reflect, pol_eff=None, **reduction_options)

    def reduce(
        self,
        spin_set_reflect,
        pol_eff=None,
        save=True,
        scale=1.0,
        **reduction_options,
    ):
        """
        Reduce a `refnx.reduce.SpinSet` of polarised neutron reflected beams,
        and correct for the efficiency of the polariser system.

        Parameters
        ----------
        spin_set_reflect :   refnx.reduce.SpinSet
            Spinset of reflected beams
        pol_eff :   refnx.reduce.PolarisationEfficiency, optional
            Input a defined polarisation efficiency of the
            polariser - flipper 1 - flipper 2 - analyser system.
        reduction_options :   dict, optional
            Reduction options to apply to every spin channel being reduced.
            This will override any individually defined reduction options
            for each spin channel

        Attributes
        ----------
        spin_set_reflect :   refnx.reduce.SpinSet
            Reflected beams from PNR experiment
        """
        # get a default set of reduction options
        options = ReductionOptions()
        options.update(reduction_options)

        # set up the wavelength bins
        if options["wavelength_bins"] is None:
            wb = calculate_wavelength_bins(
                options["lo_wavelength"],
                options["hi_wavelength"],
                options["rebin_percent"],
            )
            options["wavelength_bins"] = wb

        # a list of which datasets has been reduced ok
        self._reduced_successfully = []

        # go through each spin channel and reduce it
        for sc, reducer in self.reducers.items():
            # first get the correct reduction options
            rdo = spin_set_reflect.sc_opts[sc]
            if rdo is None:
                rdo = options
            else:
                # overwrite properties that need to be common
                rdo["wavelength_bins"] = options["wavelength_bins"]
                rdo["lo_wavelength"] = options["lo_wavelength"]
                rdo["hi_wavelength"] = options["hi_wavelength"]
                rdo["rebin_percent"] = options["rebin_percent"]

            db = reducer
            rb = getattr(spin_set_reflect, sc)
            if rb is not None:
                db.reduce(rb, save=save, scale=scale, **rdo)
                self._reduced_successfully.append(sc)
            else:
                # no reflected beam for a spin channel
                continue

            assert (
                reducer.reflected_beam.m_spec.shape
                == reducer.reflected_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec.shape
                == reducer.direct_beam.m_spec_sd.shape
            )
            assert (
                reducer.direct_beam.m_spec.shape
                == reducer.reflected_beam.m_spec.shape
            )

        # make sure that the "ud and "du" direct beams get processed. This is
        # the "spin leakage". If everything was perfect this would be 0.
        for sc in ["du", "ud"]:
            pn = self.spin_set_direct.channels[sc]
            if pn is not None:
                pn.process(**rdo)

        # by this point an unpolarised reduction has been done, but we need to
        # correct the spectra for PNR. The following spectra (N, T) should be
        # overwritten:
        # self.reducers[sc].reflected_beam.m_spec
        # self.reducers[sc].reflected_beam.m_spec_sd
        # self.reducers[sc].direct_beam.m_spec
        # self.reducers[sc].direct_beam.m_spec_sd
        # THIS IS WHERE THE MAGIC HAPPENS
        self._efficiency_correction(pol_eff=pol_eff)

        # once the wavelength spectra have been corrected/overwritten then the
        # reflectivities need to be recalculated.
        # this doesn't correct the offspecular
        for sc in self._reduced_successfully:
            reducer = self.reducers[sc]
            # Add ycorr to reducer attributes and divide
            # by the corrected reflected beams by direct beams
            reducer.y_corr, reducer.y_corr_err = EP.EPdiv(
                reducer.reflected_beam.m_spec_polcorr,
                reducer.reflected_beam.m_spec_sd,
                reducer.direct_beam.m_spec_polcorr,
                reducer.direct_beam.m_spec_sd,
            )
            # Apply scale to reduced data
            reducer.y_corr /= scale
            reducer.y_corr_err /= scale
            if save:
                # now write out the corrected reflectivity files
                fnames = []
                datasets = []
                datafilename = reducer.reflected_beam.datafilename
                datafilename = os.path.basename(
                    datafilename.split(".nx.hdf")[0]
                )

                for i in range(np.size(reducer.y_corr, 0)):
                    data = reducer.data(scanpoint=i)
                    data_tup = (
                        data[0],
                        reducer.y_corr[i],
                        reducer.y_corr_err[i],
                        data[-1],
                    )
                    datasets.append(ReflectDataset(data_tup))

                    for i, dataset in enumerate(datasets):
                        fname = f"{datafilename}_{i}_PolCorr.dat"
                        fnames.append(fname)
                        with open(fname, "wb") as f:
                            dataset.save(f)

    def _efficiency_correction(self, pol_eff=None):
        """
        Applies the combined efficiency matrix correction to raw spectra. The
        efficiency correction is given by pol_eff and should be supplied by
        `refnx.reduce.PolarisedReduce.reduce`.

        Parameters
        ----------
        reducers : dict of PlatypusReduce objects
            reducer objects for each spin channel

        pol_eff  :   optional, refnx.reduce.PolarisationEfficiency object
            Describes the polarisation efficiency of PLATYPUS. If None,
            then will initialise the standard efficiency during the
            correction process. *advanced users only*

        Returns
        ----------
        corrected_reducers  :   dict of PlatypusReduce objects
            reducer objects for each spin channel that have spectra with
            the suffix "_polcorr" that has been corrected for the
            polarisation efficiency of the PLATYPUS setup.
        """

        # If only one spin-flip channel is recorded, assume both
        # spin-flip channels are identical
        m_spec = self.reducers["dd"].reflected_beam.m_spec

        measured = set(self._reduced_successfully)
        sf = {"du", "ud"}
        nsf = {"dd", "uu"}

        # Create dict of direct and reflected beam spectra
        rb_spectra = {}
        db_spectra = {}

        for sc, reducer in self.reducers.items():
            # dd, du, ud, uu
            # NSF
            if sc in nsf:
                # these should definitely be measured
                rb_spectra[sc] = reducer.reflected_beam.m_spec
                db_spectra[sc] = reducer.direct_beam.m_spec

            if sc in sf:
                # Need to get spin-flip direct beams from SpinSet since
                # we don't include them in the reducer
                pn = self.spin_set_direct.channels[sc]
                if pn is None:
                    db_spectra[sc] = np.zeros_like(m_spec)
                else:
                    # this is the spectrum that "leaks" through when you
                    # measure a "spin flip" direct beam
                    db_spectra[sc] = self.spin_set_direct.channels[sc].m_spec

                if sc in self._reduced_successfully:
                    # you measured the spin channel
                    rb_spectra[sc] = reducer.reflected_beam.m_spec
                elif measured.intersection(sf):
                    # you don't have the spin channel, but you have the other
                    it = measured.intersection(sf).pop()
                    rb_spectra[sc] = self.reducers[it].reflected_beam.m_spec
                else:
                    # you have no SF channels
                    rb_spectra[sc] = np.zeros_like(m_spec)

        if pol_eff is None:
            # Define polarisation efficiency of PLATYPUS

            # If the analyser is out of the beam and the mode is POL, then
            # we assume that the analyser and flipper 2 have a perfect
            # efficiency. Otherwise if analyser is in the beam and the mode
            # is POLANAL, then use real efficiencies.

            # Check whether mode is POLANAL or just POL instead of this
            if self.reducers["dd"].reflected_beam.cat.mode == "POL":
                # if mode is POL then analyser is out of the beam, and config
                # only uses polariser and flipper1
                config = "PF"
            elif self.reducers["dd"].reflected_beam.cat.mode == "POLANAL":
                # if mode is POLANAL, analyser is in beam and
                # polarisation config uses all elements.
                config = "full"

            pol_eff = PolarisationEfficiency(
                self.reducers["dd"].reflected_beam.m_lambda[0], config=config
            )
        else:
            if not isinstance(pol_eff, PolarisationEfficiency):
                raise ValueError()

        # Define sizes of corrected beam spectra (N, T, 4, 1) and
        # inverted combined efficiency matrix to be (1, T, 4, 4)
        # where T is the number of wavelength bins

        # Invert and apply the refnx.reduce.PolarisationEfficiency parameters
        # to the raw spectra to correct for efficiencies.
        inverted_combined_efficiency_matrix = np.linalg.inv(
            pol_eff.combined_efficiency_matrix
        )
        # Create numpy arrays with shape (N, T, S, 1) (n_spectra, tof,
        # spin_channels, 1) to broadcast with array of
        # efficiency matrices
        N_TBINS = m_spec.shape[1]
        MAX_N_SPECTRA = np.max([s.shape[0] for s in rb_spectra.values()])

        raw_db = np.zeros([MAX_N_SPECTRA, N_TBINS, 4, 1])
        raw_rb = np.zeros([MAX_N_SPECTRA, N_TBINS, 4, 1])

        raw_db[:, :, 0, 0] = db_spectra["dd"]
        raw_db[:, :, 1, 0] = db_spectra["du"]
        raw_db[:, :, 2, 0] = db_spectra["ud"]
        raw_db[:, :, 3, 0] = db_spectra["uu"]

        raw_rb[:, :, 0, 0] = rb_spectra["dd"]
        raw_rb[:, :, 1, 0] = rb_spectra["du"]
        raw_rb[:, :, 2, 0] = rb_spectra["ud"]
        raw_rb[:, :, 3, 0] = rb_spectra["uu"]

        corrected_db = inverted_combined_efficiency_matrix @ raw_db
        corrected_rb = inverted_combined_efficiency_matrix @ raw_rb

        # Assign corrected spectra to m_spec_polcorr, and reshape to (N, T, 4).
        # TODO handle uncertainties
        for sc in self._reduced_successfully:
            # NOTE: corrected_db has the spin channels in reverse order
            # compared to raw_rb/raw_db, the I00 channel corresponds to the R++
            # channel in the matrix formulation. THIS IS THE REVERSE OF WHAT
            # I'D EXPECT, BUT SYNCS WITH THE WILDES PAPER.
            idx = ["uu", "ud", "du", "dd"].index(sc)
            reducer = self.reducers[sc]

            # TODO think about the reshape for N_SPECTRA
            reducer.direct_beam.m_spec_polcorr = corrected_db[
                :, :, idx, 0
            ].reshape(m_spec.shape)

            reducer.reflected_beam.m_spec_polcorr = corrected_rb[
                :, :, idx, 0
            ].reshape(m_spec.shape)

            # If spin-flip channel, replace direct beam with corrected
            # non-spin-flip counterpart
            if sc == "du":
                reducer.direct_beam.m_spec_polcorr = corrected_db[
                    :, :, -1, 0
                ].reshape(m_spec.shape)
            elif sc == "ud":
                reducer.direct_beam.m_spec_polcorr = corrected_db[
                    :, :, 0, 0
                ].reshape(m_spec.shape)


class SpatzReduce(ReflectReduce):
    """
    Reduces Spatz reflectometer data to give the specular reflectivity.
    Offspecular data maps are also produced.

    Parameters
    ----------
    direct : string, hdf5 file-handle or SpatzNexus object
        A string containing the path to the direct beam hdf5 file,
        the hdf5 file itself, or a SpatzNexus object.
    data_folder : str, optional
        Where is the raw data stored?

    Examples
    --------

    >>> from refnx.reduce import SpatzReduce
    >>> reducer = SpatzReduce('SPZ0000711.nx.hdf')
    >>> datasets, reduced = reducer.reduce('SPZ0000711.nx.hdf',
    ...                                    rebin_percent=2)

    """

    def __init__(self, direct, data_folder=None, **kwds):
        super().__init__(direct, "SPZ", data_folder=data_folder)

    def _reduce_single_angle(self, scale=1):
        """
        Reduce a single angle.
        """
        n_spectra = self.reflected_beam.n_spectra
        n_tpixels = np.size(self.reflected_beam.m_topandtail, 1)
        n_xpixels = np.size(self.reflected_beam.m_topandtail, 2)

        m_twotheta = np.zeros((n_spectra, n_tpixels, n_xpixels))

        detrot_difference = (
            self.reflected_beam.detector_z - self.direct_beam.detector_z
        )

        # difference in pixels between reflected position and direct beam
        # at the two different detrots.
        QZ_PIXEL_SPACING = self.reflected_beam.cat.qz_pixel_size[0]
        dy = self.reflected_beam.detector_y

        # convert that pixel difference to angle (in small angle approximation)
        # higher `som` leads to lower m_beampos. i.e. higher two theta
        # is at lower pixel values
        beampos_2theta_diff = -(
            self.reflected_beam.m_beampos - self.direct_beam.m_beampos
        )
        beampos_2theta_diff *= QZ_PIXEL_SPACING / dy[0]
        beampos_2theta_diff = np.degrees(beampos_2theta_diff)

        total_2theta_deflection = detrot_difference + beampos_2theta_diff

        # omega_nom.shape = (N, )
        omega_nom = total_2theta_deflection / 2.0
        omega_corrected = omega_nom[:, np.newaxis]

        m_twotheta += np.arange(n_xpixels * 1.0)[np.newaxis, np.newaxis, :]
        m_twotheta -= self.direct_beam.m_beampos[:, np.newaxis, np.newaxis]
        # minus sign in following line because higher two theta is at lower
        # pixel values
        m_twotheta *= -QZ_PIXEL_SPACING / dy[:, np.newaxis, np.newaxis]
        m_twotheta = np.degrees(m_twotheta)
        m_twotheta += detrot_difference

        # you may be reflecting upside down, reverse the sign.
        upside_down = np.sign(omega_corrected[:, 0])
        m_twotheta *= upside_down[:, np.newaxis, np.newaxis]
        omega_corrected *= upside_down[:, np.newaxis]

        self.omega_corrected = omega_corrected
        self.m_twotheta = m_twotheta
        self.n_spectra = n_spectra


def reduce_stitch(
    reflect_list,
    direct_list,
    data_folder=None,
    prefix="PLP",
    trim_trailing=True,
    save=True,
    scale=1.0,
    reduction_options=None,
):
    """
    Reduces a list of reflected beam run numbers and a list of corresponding
    direct beam run numbers from the Platypus/Spatz reflectometers. If there
    are multiple reflectivity files they are spliced together.

    Parameters
    ----------
    reflect_list : list
        Reflected beam run numbers, e.g. `[708, 709, 710]`
        708 corresponds to the file PLP0000708.nx.hdf.
    direct_list : list
        Direct beam run numbers, e.g. `[711, 711, 711]`
    data_folder : str, optional
        Where is the raw data stored?
    prefix : str, optional
        The instrument filename prefix.
    trim_trailing : bool, optional
        When datasets are spliced together do you want to remove points in the
        overlap region from the preceding dataset?
    save : bool, optional
        If `True` then the spliced file is written to a file (in the working
        directory) with a name like: `c_PLP0000708.dat`.
    scale : float, optional
        Scales the data by this value.
    reduction_options : None, dict, or list of dict, optional
        Options passed directly to `refnx.reduce.PlatypusNexus.process`,
        for processing of individual spectra. Look at that method docstring
        for specification of options. If an individual dict then the same
        options are used to process all datasets. A list (or sequence) of
        dict can be used to specify different options for each datasets. If
        None, then a default set of reduction options will be used.

    Returns
    -------
    combined_dataset, reduced_filename : refnx.dataset.ReflectDataset, str
        The combined dataset and the file name of the reduced data, if it was
        saved. If it wasn't saved `reduced_filename` is `None`.

    Notes
    -----
    The `prefix` is used to specify the run numbers to a filename.
    For example a run number of 10, and a prefix of `PLP` resolves to a
    NeXus filename of 'PLP0000010.nx.hdf'.

    Examples
    --------

    >>> from refnx.reduce import reduce_stitch
    >>> dataset, fname = reduce_stitch([708, 709, 710],
    ...                                [711, 711, 711],
    ...                                reduction_options={"rebin_percent": 2})

    """
    options = [ReductionOptions()] * len(reflect_list)
    try:
        if reduction_options is not None:
            options = []
            for i in range(len(reflect_list)):
                if isinstance(reduction_options[i], dict):
                    options.append(reduction_options[i])
                else:
                    options.append(ReductionOptions())
    except KeyError:
        # reduction_options may be an individual dict
        if isinstance(reduction_options, dict):
            options = [reduction_options] * len(reflect_list)

    # now reduce all the files.
    zipped = zip(reflect_list, direct_list, options)

    combined_dataset = ReflectDataset()

    if data_folder is None:
        data_folder = os.getcwd()

    if prefix == "PLP":
        reducer_klass = PlatypusReduce
    elif prefix == "SPZ":
        reducer_klass = SpatzReduce
    else:
        raise ValueError("Incorrect prefix specified")

    for index, val in enumerate(zipped):
        reflect_datafile = os.path.join(
            data_folder, number_datafile(val[0], prefix=prefix)
        )
        direct_datafile = os.path.join(
            data_folder, number_datafile(val[1], prefix=prefix)
        )

        reducer = reducer_klass(direct_datafile)
        datasets, fnames = reducer.reduce(
            reflect_datafile, save=save, **val[2]
        )

        if not index:
            datasets[0].scale(scale)

        combined_dataset.add_data(
            datasets[0].data, requires_splice=True, trim_trailing=trim_trailing
        )

    fname_dat = None

    if save:
        # this will give us <fname>.nx.hdf
        # if reflect_list was an integer you'll get PLP0000708.nx.hdf
        fname = number_datafile(reflect_list[0], prefix=prefix)
        # now chop off .nx.hdf extension
        fname = basename_datafile(fname)

        fname_dat = "c_{0}.dat".format(fname)
        with open(fname_dat, "wb") as f:
            combined_dataset.save(f)
        fname_xml = "c_{0}.xml".format(fname)
        with open(fname_xml, "wb") as f:
            combined_dataset.save_xml(f)

    return combined_dataset, fname_dat


class AutoReducer:
    """
    Auto-reduces reflectometry data.

    Watches a datafolder for new/modified NeXUS files and reduces them.

    Parameters
    ----------
    direct_beams: list of {str, h5data}
        list of str, or list of h5py file handles pointing to
        direct beam runs

    scale: float or array-like
        Scale factors corresponding to each direct beam.

    reduction_options: dict, or list of dict
        Specifies the reduction options for each of the direct beams.
        A default set of options is provided by
        `refnx.reduce.ReductionOptions`.

    data_folder: {str, Path}
        Path to the data folder containing the data to be reduced.

    Notes
    -----
    Requires that the 'watchdog' package be installed. Starts two threads that
    are responsible for doing the reduction.
    """

    def __init__(
        self, direct_beams, scale=1, reduction_options=None, data_folder="."
    ):
        from watchdog.observers import Observer
        from refnx.reduce._auto_reduction import NXEH

        self.data_folder = data_folder

        # deal with reduction options first
        options = [ReductionOptions()] * len(direct_beams)
        try:
            if reduction_options is not None:
                options = []
                for i in range(len(direct_beams)):
                    if isinstance(reduction_options[i], dict):
                        options.append(reduction_options[i])
                    else:
                        options.append(ReductionOptions())
        except KeyError:
            # reduction_options may be an individual dict
            if isinstance(reduction_options, dict):
                options = [reduction_options] * len(direct_beams)

        # deal with scale factors
        scale_factors = np.broadcast_to(scale, len(direct_beams))
        scale_factors = scale_factors.astype(np.float)

        zipped = zip(direct_beams, options, scale_factors)

        # work out what type of instrument you have
        d0 = direct_beams[0]
        self.redn_klass = PlatypusReduce
        self.reflect_klass = PlatypusNexus
        if (isinstance(d0, str) and d0.startswith("SPZ")) or (
            isinstance(d0, h5py.File) and d0.filename.startswith("SPZ")
        ):
            self.redn_klass = SpatzReduce
            self.reflect_klass = SpatzNexus

        self.direct_beams = {}
        for direct_beam, ro, scale_factor in zipped:
            rn = self.reflect_klass(direct_beam)
            db = self.redn_klass(rn)
            fname = os.path.basename(rn.cat.filename)
            self.direct_beams[fname] = {
                "reflectnexus": rn,
                "reducer": db,
                "collimation": np.r_[rn.cat.ss_coll1, rn.cat.ss_coll2],
                "reduction_options": ro,
                "scale": scale_factor,
            }

        self.redn_cache = {}
        self._redn_cache_tbl = pd.DataFrame(
            columns=["fname", "sample_name", "omega"]
        )

        # start watching the data_folder
        self.queue = Queue()

        event_handler = NXEH(self.queue)
        observer = Observer()
        observer.schedule(event_handler, path=self.data_folder)
        observer.start()

        self.worker = Thread(target=self)
        self.worker.setDaemon(True)
        self.worker.start()

    def __call__(self):
        while True:
            if not self.queue.empty():
                # file might still be being written by SICS? allow a bit of
                # time for it to complete.
                time.sleep(1.5)

                event = self.queue.get()
                # print(event.src_path)
                rb = self.reflect_klass(event.src_path)
                fname = os.path.basename(rb.cat.filename)
                db = self.match_direct_beam(rb)

                if db is not None:
                    # the reduction
                    entry = self.direct_beams[db]
                    reducer = entry["reducer"]
                    opts = entry["reduction_options"]
                    scale = entry["scale"]
                    try:
                        datasets, _ = reducer.reduce(rb, scale=scale, **opts)
                    except Exception as e:
                        # don't want to stop reducing if there is an error
                        # somewhere
                        print(e)
                        continue

                    print(f"Reduced: {fname}")

                    for i, dataset in enumerate(datasets):
                        dataset.filename = f"{fname.rstrip('.nx.hdf')}_{i}.dat"

                    # save the reduced files in a cache
                    sample_name = rb.cat.sample_name.tobytes()
                    sample_name = sample_name.decode("utf-8")[:-1]

                    omega = float(rb.cat.omega[0])
                    self.redn_cache[fname] = {
                        "datasets": datasets,
                        "sample_name": sample_name,
                        "omega": omega,
                    }

                    data = {
                        "fname": [fname],
                        "sample_name": [sample_name],
                        "omega": omega,
                    }
                    if not self._redn_cache_tbl.fname.str.match(fname).any():
                        # see if the entry is already in the _redn_cache_tbl
                        # if it is, then you don't want to add it again
                        entry = pd.DataFrame(data=data)
                        self._redn_cache_tbl = self._redn_cache_tbl.append(
                            entry
                        )

                    # now splice matching datasets
                    ds = self.match_datasets(self.redn_cache[fname])
                    if len(ds) > 1:
                        try:
                            c = self.splice_datasets(ds)
                        except Exception as e:
                            print(e)
                            continue
                        print(
                            f"Combined into: {c}, {[d.filename for d in ds]}"
                        )
            else:
                time.sleep(5.0)

    def match_direct_beam(self, rb):
        """
        Finds the direct beam associated with a reflection measurement.
        Matching is done by finding identical collimation conditions.

        Parameters
        ----------
        rb: {PlatypusNexus, SpatzNexus}
            The reflectometry run.

        Returns
        -------
        db: str
            The direct beam file name that matches the reflection measurement
            This is used to look up an entry in `AutoReducer.direct_beams`.
        """
        # the reflected-direct beam match is done via slit sizes
        # (not infallible)
        collimation = np.r_[rb.cat.ss_coll1, rb.cat.ss_coll2]
        for k, v in self.direct_beams.items():
            if np.allclose(collimation, v["collimation"], atol=0.01):
                return k
        return None

    def match_datasets(self, dct):
        """
        Finds all the datasets in `AutoReducer.redn_cache` that share an
        *identical* `dct["sample_name"]`, but may have been measured at
        different angles.

        Parameters
        ----------
        dct: dict
            dct.keys() = ["fname", "sample_name", "omega"]

        Returns
        -------
        datasets: list of `refnx.dataset.Data1D`
            Datasets that share the same sample_name as `dct['sample_name']`
        """
        tbl = self._redn_cache_tbl
        # exact match
        fnames = tbl.fname[
            tbl.sample_name.str.match(f"^{dct['sample_name']}$")
        ]
        datasets = [self.redn_cache[fname]["datasets"][0] for fname in fnames]

        return datasets

    def splice_datasets(self, ds):
        """
        Combines datasets together.

        Parameters
        ----------
        ds: list of `refnx.dataset.Data1D`
            The datasets to splice together.

        Returns
        -------
        fname: str
            The name of the combined dataset

        Notes
        -----
        The combined dataset is saved as `f"c_{d.filename}.dat"`,
        where d is the dataset with the lowest average Q value
        from ds.
        """
        appended_ds = ReflectDataset()

        datasets = []
        average_q = []
        for d in ds:
            dataset = ReflectDataset(d)
            average_q.append(np.mean(dataset.x))
            datasets.append(dataset)

        idxs = np.argsort(average_q)
        # sort datasets according to average Q.
        datasets = [d for _, d in sorted(zip(idxs, datasets))]

        for dataset in datasets:
            appended_ds += dataset

        fname = datasets[0].filename.rstrip(".dat")
        fname = fname.split("_")[0]
        fname = f"c_{fname}.dat"
        appended_ds.save(fname)
        return fname


if __name__ == "__main__":
    print(strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime()))

    a = reduce_stitch([708, 709, 710], [711, 711, 711], rebin_percent=2)

    a.save("test1.dat")

    print(strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime()))



import refnx.util.general as general
import refnx
import numpy as np
import os
from numpy.testing import assert_almost_equal, assert_, assert_allclose


def test_version():
    # check that we can retrieve a version string
    refnx.__version__


class TestGeneral:
    def setup_method(self):
        self.pth = os.path.dirname(os.path.abspath(__file__))

    def test_q(self):
        q = general.q(1.0, 2.0)
        assert_almost_equal(q, 0.1096567037)

    def test_q2(self):
        qx, qy, qz = general.q2(1.0, 2.0, 0.0, 2.0)
        assert_almost_equal(qz, 0.1096567037)

    def test_wavelength_velocity(self):
        speed = general.wavelength_velocity(20.0)
        assert_almost_equal(speed, 197.8017006541796, 5)

    def test_wavelength(self):
        wavelength = general.wavelength(0.1096567037, 1.0)
        assert_almost_equal(wavelength, 2.0)

    def test_angle(self):
        angle = general.angle(0.1096567037, 2.0)
        assert_almost_equal(angle, 1.0)

    def test_dict_compare(self):
        c = {"f": np.arange(10)}
        d = {"f": np.arange(10)}

        assert_(general._dict_compare(c, d))

        d = {"f": np.arange(11)}
        assert_(not general._dict_compare(c, d))

        d = {"f": 2}
        assert_(not general._dict_compare(c, d))

        assert_(general._dict_compare({"a": 1}, {"a": 1}))

    def test_neutron_transmission(self):
        try:
            import periodictable as pt
        except ImportError:
            return

        mat = pt.formula("N2", density=1.25e-3)
        ntd = general._neutron_transmission_depth(mat, 2)
        assert_allclose(ntd, 1365.8010284973458 * 10)

        t = general.neutron_transmission(
            "N2", 1.25e-3, 2, 1365.8010284973458 * 10
        )
        assert_almost_equal(t, np.exp(-1.0))

        # check that we can vectorise
        t = general.neutron_transmission(
            "N2", 1.25e-3, [2, 2.0], [1365.8010284973458 * 10] * 2
        )
        assert_almost_equal(t, [np.exp(-1.0)] * 2)



import os.path
from PyQt5 import QtWidgets, uic
from PyQt5.QtCore import pyqtSlot
import numpy as np


UI_LOCATION = os.path.join(os.path.dirname(os.path.abspath(__file__)), "ui")


class OptimisationParameterView(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.ui = uic.loadUi(
            os.path.join(UI_LOCATION, "optimisation.ui"), self
        )

    def parameters(self, method):
        """
        Parameters
        ----------
        method : {'differential_evolution'}
            The fit method

        Returns
        -------
        opt_par : dict
            Options for fitting algorithm.
        """
        kws = {}
        if method == "DE":
            kws["strategy"] = self.ui.de_strategy.currentText()
            kws["maxiter"] = self.ui.de_maxiter.value()
            kws["popsize"] = self.ui.de_popsize.value()
            kws["tol"] = self.ui.de_tol.value()
            kws["atol"] = self.ui.de_atol.value()
            kws["init"] = self.ui.de_initialisation.currentText()
            kws["recombination"] = self.ui.de_recombination.value()
            mutation_lb = self.ui.de_mutation_lb.value()
            mutation_ub = self.ui.de_mutation_ub.value()
            kws["mutation"] = (
                min(mutation_lb, mutation_ub),
                max(mutation_lb, mutation_ub),
            )
            target = self.ui.de_target.currentText()
            if target == "log-posterior":
                target = "nlpost"
            else:
                target = "nll"
            kws["target"] = target
        elif method == "dual_annealing":
            kws["maxiter"] = self.ui.da_maxiter.value()
            kws["initial_temp"] = self.ui.da_initial_temp.value()
            kws["restart_temp_ratio"] = self.ui.da_restart_temp.value()
            kws["visit"] = self.ui.da_visit.value()
            kws["accept"] = self.ui.da_accept.value()
            kws["no_local_search"] = (
                self.ui.da_no_local_search.isChecked() is True
            )

            target = self.ui.de_target.currentText()
            if target == "log-posterior":
                target = "nlpost"
            else:
                target = "nll"
            kws["target"] = target
        elif method == "L-BFGS-B":
            kws["maxiter"] = self.ui.lbfgsb_maxiter.value()
            target = self.ui.lbfgsb_target.currentText()
            if target == "log-posterior":
                target = "nlpost"
            else:
                target = "nll"
            kws["target"] = target
        elif method == "SHGO":
            target = self.ui.shgo_target.currentText()
            kws["n"] = self.ui.shgo_n.value()
            kws["iters"] = self.ui.shgo_iters.value()
            if target == "log-posterior":
                target = "nlpost"
            else:
                target = "nll"
            kws["target"] = target

        return kws



"""
GraphProperties: storing information about Matplotlib traces
e.g. colors, linesizes, etc.
"""
import matplotlib.artist as artist


_requiredgraphproperties = {
    "lw": float,
    "label": str,
    "linestyle": str,
    "fillstyle": str,
    "marker": str,
    "markersize": float,
    "markeredgecolor": str,
    "markerfacecolor": str,
    "zorder": int,
    "color": str,
    "visible": bool,
}

_lines = {
    "data_properties": dict,
    "fit_properties": dict,
    "sld_profile_properties": dict,
    "residuals_properties": dict,
}


class GraphProperties(dict):
    def __init__(self):
        self["visible"] = True

        # a Matplotlib trace
        self["ax_data"] = None
        self["ax_fit"] = None
        self["ax_residuals"] = None
        self["ax_sld_profile"] = None
        self["data_properties"] = {}
        self["fit_properties"] = {}
        self["sld_profile_properties"] = {}
        self["residuals_properties"] = {}

    def __getattr__(self, key):
        if key in _requiredgraphproperties:
            return self[key]

        try:
            v = self[key]
        except KeyError:
            return None

        return v

    def __setstate__(self, state):
        self["visible"] = state["visible"]
        self["ax_data"] = None
        self["ax_fit"] = None
        self["ax_sld_profile"] = None
        for line in _lines:
            self[line] = state[line]

    def __getstate__(self):
        self.save_graph_properties()
        d = {}
        d["ax_data"] = None
        d["ax_fit"] = None
        d["ax_sld_profile"] = None
        d["visible"] = self["visible"]
        for line in _lines:
            d[line] = self[line]
        return d

    def save_graph_properties(self):
        # pass
        if self.ax_data is not None:
            for key in _requiredgraphproperties:
                # ax_data is an ErrorbarContainer
                self["data_properties"][key] = artist.getp(
                    self.ax_data[0], key
                )

        if self.ax_fit is not None:
            for key in _requiredgraphproperties:
                self["fit_properties"][key] = artist.getp(self.ax_fit, key)

        if self.ax_residuals is not None:
            for key in _requiredgraphproperties:
                self["residuals_properties"][key] = artist.getp(
                    self.ax_residuals, key
                )

        if self.ax_sld_profile is not None:
            for key in _requiredgraphproperties:
                self["sld_profile_properties"][key] = artist.getp(
                    self.ax_sld_profile, key
                )



"""
Component for studying lipid membranes at an interface
"""

import numpy as np
from refnx.reflect import Component, SLD, ReflectModel, Structure
from refnx.analysis import possibly_create_parameter, Parameters, Parameter
from refnx.reflect.structure import overall_sld


class LipidLeaflet(Component):
    r"""
    Describes a lipid leaflet Component at an interface

    Parameters
    ----------
    APM: float or refnx.analysis.Parameter
    b_heads: float, refnx.analysis.Parameter, complex or SLD
        Sum of coherent scattering lengths of head group (Angstrom).
        When an SLD is provided it is simply an easy way to provide a complex
        value. `LipidLeaflet.b_heads_real` is set to `SLD.real`, etc.
    vm_heads: float or refnx.analysis.Parameter
        Molecular volume of head group (Angstrom**2)
    thickness_heads: float or refnx.analysis.Parameter
        Thickness of head group region (Angstrom)
    b_tails: float, refnx.analysis.Parameter, complex or SLD
        Sum of coherent scattering lengths of tail group (Angstrom).
        When an SLD is provided it is simply an easy way to provide a complex
        value. `LipidLeaflet.b_tails_real` is set to `SLD.real`, etc.
    vm_tails: float or refnx.analysis.Parameter
        Molecular volume of tail group (Angstrom**2)
    thickness_tails: float or refnx.analysis.Parameter
        Thickness of head group region (Angstrom)
    rough_head_tail: float or refnx.analysis.Parameter
        Roughness of head-tail group (Angstrom)
    rough_preceding_mono: float or refnx.analysis.Parameter
        Roughness between preceding component (in the fronting direction) and
        the monolayer (Angstrom). If `reverse_monolayer is False` then this is
        the roughness between the preceding component and the heads, if
        `reverse_monolayer is True` then this is the roughness between the
        preceding component and the tails.
    head_solvent: None, float, complex, refnx.reflect.SLD
        Solvent for the head region. If `None`, then solvation will be
        performed by the parent `Structure`, using the `Structure.solvent`
        attribute. Other options are coerced to an `SLD` object using
        `SLD(float | complex)`. A float/complex argument is the SLD of the
        solvent (10**-6 Angstrom**-2).
    tail_solvent: None, float, complex, refnx.reflect.SLD
        Solvent for the tail region. If `None`, then solvation will be
        performed by the parent `Structure`, using the `Structure.solvent`
        attribute. Other options are coerced to an `SLD` object using
        `SLD(float | complex)`. A float/complex argument is the SLD of the
        solvent (10**-6 Angstrom**-2).
    reverse_monolayer: bool, optional
        The default is to have heads closer to the fronting medium and
        tails closer to the backing medium. If `reverse_monolayer is True`
        then the tails will be closer to the fronting medium and heads
        closer to the backing medium.
    name: str, optional
        The name for the component

    Notes
    -----
    The sum of coherent scattering lengths must be in Angstroms, the volume
    must be in cubic Angstroms. This is because the SLD of a tail group is
    calculated as `b_tails / vm_tails * 1e6` to achieve the units
    10**6 Angstrom**-2.
    """

    # TODO: use SLD of head instead of b_heads, vm_heads?
    def __init__(
        self,
        apm,
        b_heads,
        vm_heads,
        thickness_heads,
        b_tails,
        vm_tails,
        thickness_tails,
        rough_head_tail,
        rough_preceding_mono,
        head_solvent=None,
        tail_solvent=None,
        reverse_monolayer=False,
        name="",
    ):
        """
        Parameters
        ----------
        apm: float or Parameter
            Area per molecule
        b_heads: float, Parameter or complex
            Sum of coherent scattering lengths of head group (Angstrom)
        vm_heads: float or Parameter
            Molecular volume of head group (Angstrom**3)
        thickness_heads: float or Parameter
            Thickness of head group region (Angstrom)
        b_tails: float, Parameter or complex
            Sum of coherent scattering lengths of tail group (Angstrom)
        vm_tails: float or Parameter
            Molecular volume of tail group (Angstrom**3)
        thickness_tails: float or Parameter
            Thickness of head group region (Angstrom)
        rough_head_tail: float or refnx.analysis.Parameter
            Roughness of head-tail group (Angstrom)
        rough_preceding_mono: float or Parameter
            Roughness between preceding component (in the fronting direction)
            and the monolayer (Angstrom). If `reverse_monolayer is False` then
            this is the roughness between the preceding component and the
            heads, if `reverse_monolayer is True` then this is the roughness
            between the preceding component and the tails.
        head_solvent: None, float, complex, SLD
            Solvent for the head region. If `None`, then solvation will be
            performed by the parent `Structure`, using the `Structure.solvent`
            attribute. Other options are coerced to an `SLD` object using
            `SLD(float | complex)`. A float/complex argument is the SLD of the
            solvent (10**-6 Angstrom**-2).
        tail_solvent: None, float, complex, SLD
            Solvent for the tail region. If `None`, then solvation will be
            performed by the parent `Structure`, using the `Structure.solvent`
            attribute. Other options are coerced to an `SLD` object using
            `SLD(float | complex)`. A float/complex argument is the SLD of the
            solvent (10**-6 Angstrom**-2).
        reverse_monolayer: bool, optional
            The default is to have heads closer to the fronting medium and
            tails closer to the backing medium. If `reverse_monolayer is True`
            then the tails will be closer to the fronting medium and heads
            closer to the backing medium.
        name: str, optional
            The name for the component
        """
        super().__init__()
        self.apm = possibly_create_parameter(
            apm, "%s - area_per_molecule" % name, units="Å**2"
        )

        if isinstance(b_heads, complex):
            self.b_heads_real = possibly_create_parameter(
                b_heads.real, name="%s - b_heads_real" % name
            )
            self.b_heads_imag = possibly_create_parameter(
                b_heads.imag, name="%s - b_heads_imag" % name
            )
        elif isinstance(b_heads, SLD):
            self.b_heads_real = b_heads.real
            self.b_heads_imag = b_heads.imag
        else:
            self.b_heads_real = possibly_create_parameter(
                b_heads, name="%s - b_heads_real" % name
            )
            self.b_heads_imag = possibly_create_parameter(
                0, name="%s - b_heads_imag" % name
            )

        self.b_heads_real.units = self.b_heads_imag.units = "Å"

        self.vm_heads = possibly_create_parameter(
            vm_heads, name="%s - vm_heads" % name, units="Å**3"
        )

        self.thickness_heads = possibly_create_parameter(
            thickness_heads, name="%s - thickness_heads" % name, units="Å"
        )

        if isinstance(b_tails, complex):
            self.b_tails_real = possibly_create_parameter(
                b_tails.real, name="%s - b_tails_real" % name
            )
            self.b_tails_imag = possibly_create_parameter(
                b_tails.imag, name="%s - b_tails_imag" % name
            )
        elif isinstance(b_tails, SLD):
            self.b_tails_real = b_tails.real
            self.b_tails_imag = b_tails.imag
        else:
            self.b_tails_real = possibly_create_parameter(
                b_tails, name="%s - b_tails_real" % name
            )
            self.b_tails_imag = possibly_create_parameter(
                0, name="%s - b_tails_imag" % name
            )
        self.b_tails_real.units = self.b_tails_imag.units = "Å"

        self.vm_tails = possibly_create_parameter(
            vm_tails, name="%s - vm_tails" % name, units="Å**3"
        )
        self.thickness_tails = possibly_create_parameter(
            thickness_tails, name="%s - thickness_tails" % name, units="Å"
        )
        self.rough_head_tail = possibly_create_parameter(
            rough_head_tail, name="%s - rough_head_tail" % name, units="Å"
        )
        self.rough_preceding_mono = possibly_create_parameter(
            rough_preceding_mono,
            name="%s - rough_fronting_mono" % name,
            units="Å",
        )

        self.head_solvent = self.tail_solvent = None
        if head_solvent is not None:
            self.head_solvent = SLD(head_solvent)
        if tail_solvent is not None:
            self.tail_solvent = SLD(tail_solvent)

        self.reverse_monolayer = reverse_monolayer
        self.name = name

    def __repr__(self):
        d = {}
        d.update(self.__dict__)
        sld_bh = SLD([self.b_heads_real, self.b_heads_imag])
        sld_bt = SLD([self.b_tails_real, self.b_tails_imag])
        d["bh"] = sld_bh
        d["bt"] = sld_bt

        s = (
            "LipidLeaflet({apm!r}, {bh!r}, {vm_heads!r}, {thickness_heads!r},"
            " {bt!r}, {vm_tails!r}, {thickness_tails!r}, {rough_head_tail!r},"
            " {rough_preceding_mono!r}, head_solvent={head_solvent!r},"
            " tail_solvent={tail_solvent!r},"
            " reverse_monolayer={reverse_monolayer}, name={name!r})"
        )
        return s.format(**d)

    def slabs(self, structure=None):
        """
        Slab representation of monolayer, as an array

        Parameters
        ----------
        structure : refnx.reflect.Structure
            The Structure hosting this Component
        """
        layers = np.zeros((2, 5))

        # thicknesses
        layers[0, 0] = float(self.thickness_heads)
        layers[1, 0] = float(self.thickness_tails)

        # real and imag SLD's
        layers[0, 1] = float(self.b_heads_real) / float(self.vm_heads) * 1.0e6
        layers[0, 2] = float(self.b_heads_imag) / float(self.vm_heads) * 1.0e6

        layers[1, 1] = float(self.b_tails_real) / float(self.vm_tails) * 1.0e6
        layers[1, 2] = float(self.b_tails_imag) / float(self.vm_tails) * 1.0e6

        # roughnesses
        layers[0, 3] = float(self.rough_preceding_mono)
        layers[1, 3] = float(self.rough_head_tail)

        # volume fractions
        # head region
        volfrac = self.vm_heads.value / (
            self.apm.value * self.thickness_heads.value
        )
        layers[0, 4] = 1 - volfrac

        if self.head_solvent is not None:
            _head_solvent = self.head_solvent.complex(
                getattr(structure, "wavelength", None)
            )
            # we do the solvation here, not in Structure.slabs
            layers[0] = overall_sld(layers[0], _head_solvent)
            layers[0, 4] = 0

        # tail region
        volfrac = self.vm_tails.value / (
            self.apm.value * self.thickness_tails.value
        )

        layers[1, 4] = 1 - volfrac
        if self.tail_solvent is not None:
            _tail_solvent = self.tail_solvent.complex(
                getattr(structure, "wavelength", None)
            )
            # we do the solvation here, not in Structure.slabs
            layers[1] = overall_sld(layers[1], _tail_solvent)
            layers[1, 4] = 0

        if self.reverse_monolayer:
            layers = np.flipud(layers)
            layers[:, 3] = layers[::-1, 3]

        return layers

    @property
    def parameters(self):
        p = Parameters(name=self.name)
        p.extend(
            [
                self.apm,
                self.b_heads_real,
                self.b_heads_imag,
                self.vm_heads,
                self.thickness_heads,
                self.b_tails_real,
                self.b_tails_imag,
                self.vm_tails,
                self.thickness_tails,
                self.rough_head_tail,
                self.rough_preceding_mono,
            ]
        )
        if self.head_solvent is not None:
            p.append(self.head_solvent.parameters)
        if self.tail_solvent is not None:
            p.append(self.tail_solvent.parameters)

        return p

    def logp(self):
        # penalise unphysical volume fractions.
        volfrac_h = self.vm_heads.value / (
            self.apm.value * self.thickness_heads.value
        )

        # tail region
        volfrac_t = self.vm_tails.value / (
            self.apm.value * self.thickness_tails.value
        )

        if volfrac_h > 1 or volfrac_t > 1:
            return -np.inf

        return 0



# -*- coding: utf-8 -*-

import numpy as np

from ..state import State
from .move import Move

__all__ = ["MHMove"]


class MHMove(Move):
    r"""A general Metropolis-Hastings proposal

    Concrete implementations can be made by providing a ``proposal_function``
    argument that implements the proposal as described below.
    For standard Gaussian Metropolis moves, :class:`moves.GaussianMove` can be
    used.

    Args:
        proposal_function: The proposal function. It should take 2 arguments: a
            numpy-compatible random number generator and a ``(K, ndim)`` list
            of coordinate vectors. This function should return the proposed
            position and the log-ratio of the proposal probabilities
            (:math:`\ln q(x;\,x^\prime) - \ln q(x^\prime;\,x)` where
            :math:`x^\prime` is the proposed coordinate).
        ndim (Optional[int]): If this proposal is only valid for a specific
            dimension of parameter space, set that here.

    """

    def __init__(self, proposal_function, ndim=None):
        self.ndim = ndim
        self.get_proposal = proposal_function

    def propose(self, model, state):
        """Use the move to generate a proposal and compute the acceptance

        Args:
            coords: The initial coordinates of the walkers.
            log_probs: The initial log probabilities of the walkers.
            log_prob_fn: A function that computes the log probabilities for a
                subset of walkers.
            random: A numpy-compatible random number state.

        """
        # Check to make sure that the dimensions match.
        nwalkers, ndim = state.coords.shape
        if self.ndim is not None and self.ndim != ndim:
            raise ValueError("Dimension mismatch in proposal")

        # Get the move-specific proposal.
        q, factors = self.get_proposal(state.coords, model.random)

        # Compute the lnprobs of the proposed position.
        new_log_probs, new_blobs = model.compute_log_prob_fn(q)

        # Loop over the walkers and update them accordingly.
        lnpdiff = new_log_probs - state.log_prob + factors
        accepted = np.log(model.random.rand(nwalkers)) < lnpdiff

        # Update the parameters
        new_state = State(q, log_prob=new_log_probs, blobs=new_blobs)
        state = self.update(state, new_state, accepted)

        return state, accepted



from os.path import join as pjoin
import os
import warnings
from collections import namedtuple

import numpy as np
import h5py
import pytest

from numpy.testing import assert_equal
import refnx.reduce.event as event
from refnx.reduce import PlatypusNexus

try:
    import refnx.reduce._cevent as _cevent
except ImportError:
    HAVE_CEVENTS = False
else:
    HAVE_CEVENTS = True


@pytest.fixture(scope="class")
def event_setup(data_directory):
    if data_directory is None:
        return None

    Event_Setup = namedtuple(
        "Event_Setup",
        [
            "data_directory",
            "event_file_path",
            "event_list",
            "fpos",
            "f",
            "t",
            "y",
            "x",
        ],
    )

    event_file_path = os.path.join(
        data_directory,
        "reduce",
        "DAQ_2012-01-20T21-50-32",
        "DATASET_0",
        "EOS.bin",
    )
    with open(event_file_path, "rb") as f:
        event_list, fpos = _cevent._cevents(f)

    f, t, y, x = event_list

    stp = Event_Setup(
        data_directory, event_file_path, event_list, fpos, f, t, y, x
    )
    return stp


class TestEvent:
    @pytest.mark.usefixtures("no_data_directory")
    def test_events_smoke(self, event_setup):
        # check that the event.events function works
        event.events(event_setup.event_file_path)

    @pytest.mark.usefixtures("no_data_directory")
    def test_num_events(self, event_setup):
        assert_equal(event_setup.x.size, 783982)

    @pytest.mark.usefixtures("no_data_directory")
    def test_max_frames(self, event_setup):
        # test reading only a certain number of frames

        if HAVE_CEVENTS:
            with open(event_setup.event_file_path, "rb") as g:
                event_list, fpos = _cevent._cevents(g, max_frames=10)
                cyf, cyt, cyy, cyx = event_list

            max_f = np.max(cyf)
            assert_equal(9, max_f)

            event_list, fpos = _cevent._cevents(
                event_setup.event_file_path, max_frames=10
            )
            cyf, cyt, cyy, cyx = event_list

            max_f = np.max(cyf)
            assert_equal(9, max_f)

    @pytest.mark.usefixtures("no_data_directory")
    def test_event_same_as_detector(self, event_setup):
        # the detector file should be the same as the event file
        # warnings filter for pixel size
        pth = pjoin(event_setup.data_directory, "reduce")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            orig_file = PlatypusNexus(pjoin(pth, "PLP0011641.nx.hdf"))

        orig_det = orig_file.cat.detector
        frames = [np.arange(0, 501744)]
        event_det, fb = event.process_event_stream(
            event_setup.event_list,
            frames,
            orig_file.cat.t_bins,
            orig_file.cat.y_bins,
            orig_file.cat.x_bins,
        )
        assert_equal(event_det, orig_det)

        # PlatypusNexus.process_event_stream should be the same as well
        det, fc, bm = orig_file.process_event_stream(frame_bins=[])
        assert_equal(det, orig_det)

    @pytest.mark.usefixtures("no_data_directory")
    def test_open_with_path(self, event_setup):
        # give the event reader a file path
        event_list, fpos = _cevent._cevents(
            event_setup.event_file_path, max_frames=10
        )
        f, t, y, x = event_list
        max_f = np.max(f)
        assert_equal(9, max_f)

    @pytest.mark.usefixtures("no_data_directory")
    def test_values(self, event_setup):
        # We know the values of all the events in the file from another program
        # test that a set of random events are correct.
        assert_equal(event_setup.f[0], 0)
        assert_equal(event_setup.t[0], 18912)
        assert_equal(event_setup.y[0], 18)
        assert_equal(event_setup.x[0], 19)

        assert_equal(event_setup.f[-1], 501743)
        assert_equal(event_setup.t[-1], 20786)
        assert_equal(event_setup.y[-1], 16)
        assert_equal(event_setup.x[-1], 13)

    @pytest.mark.usefixtures("no_data_directory")
    def test_process_event_stream(self, event_setup):
        x_bins = np.array([60.5, -60.5])
        y_bins = np.linspace(110.5, -110.5, 222)
        t_bins = np.linspace(0, 50000, 1001)

        frames = event.framebins_to_frames(np.linspace(0, 24000, 7))
        detector, fbins = event.process_event_stream(
            event_setup.event_list, frames, t_bins, y_bins, x_bins
        )
        assert_equal(detector[0, 300, 97, 0], 7)
        assert_equal(detector[1, 402, 93, 0], 1)
        assert_equal(detector[4, 509, 97, 0], 5)

        x_bins = np.array([210.5, -210.5])
        y_bins = np.linspace(110.5, -110.5, 222)
        t_bins = np.linspace(0, 50000, 1001)
        frames = [[6, 10, 12, 1000, 1001, 1002, 1003]]
        detector, fbins = event.process_event_stream(
            event_setup.event_list, frames, t_bins, y_bins, x_bins
        )
        assert_equal(np.sum(detector.ravel()), 11)

        # # now see what happens if we go too far with the frame_bins
        # frames = event.framebins_to_frames([0, 24000, 30000])
        # detector, fbins = event.process_event_stream(self.event_list,
        #                                              frames,
        #                                              t_bins,
        #                                              y_bins,
        #                                              x_bins)
        # assert_equal(np.size(detector, 0), 1)

    @pytest.mark.usefixtures("no_data_directory")
    def test_monobloc_events(self, event_setup):
        # the event file changed when the ILL monobloc detector was installed
        pth = pjoin(event_setup.data_directory, "reduce")

        event_file_path = pjoin(
            pth, "DAQ_2019-11-25T12-25-07", "DATASET_0", "EOS.bin"
        )
        data = event.events(event_file_path)
        f, t, y, x = data[0]
        assert len(t) == 71223

        x_bins = np.array([2.5, 28.5])
        y_bins = np.linspace(-0.5, 1023.5, 1025)
        t_bins = np.linspace(0, 40000, 1001)
        frames = [np.arange(0, 57599, 1)]

        detector, fbins = event.process_event_stream(
            data[0], frames, t_bins, y_bins, x_bins
        )
        detector = np.squeeze(detector)

        with h5py.File(pjoin(pth, "PLP0046853.nx.hdf"), "r") as g:
            det = np.copy(g["entry1/data/hmm"])
            det = np.squeeze(det)
            assert det.shape == (1000, 1024)

        assert_equal(np.sum(detector, 1), np.sum(det, 1))
